<?php

namespace Intergo\Wallet;

use Intergo\Wallet\Enums\BalanceEnum;
use Intergo\Wallet\Enums\WalletEnum;
use Intergo\Wallet\Exceptions\InvalidAmountException;
use Intergo\Wallet\Exceptions\InvalidBalanceException;
use Intergo\Wallet\ValueObjects\TransactionInfoVO;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Redis;

abstract class AbstractWalletService
{
    private Redis $redis;
    private DB $db;

    /**
     * AbstractWalletService constructor.
     */
    public function __construct(Redis $redis, DB $db)
    {
        $this->redis = $redis;
        $this->db = $db;
    }

    abstract public function log($message, TransactionInfoVO $transactionInfo);

    /**
     * Add Balance to user on Redis
     *
     * @param $userId
     * @param $amount
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     * @throws InvalidBalanceException
     */
    public function addBalance($userId, $amount, array $information = [])
    {
        $this->validateAmount($amount);
        $transactionInfo = new TransactionInfoVO($userId, $amount, $information);
        return $this->adjustBalance($transactionInfo);
    }

    /**
     * Add EscrowBalance to user on Redis
     *
     * @param $userId
     * @param $amount
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     */
    public function addEscrowBalance($userId, $amount, array $information = [])
    {
        $this->validateAmount($amount);
        $transactionInfo = new TransactionInfoVO($userId, $amount, $information);
        $escrowBalance = $transactionInfo->setEscrowBalance($this->getEscrowBalance($userId));
        $balance = $transactionInfo->setBalance($this->getBalance($userId));
        if (($escrowBalance + $amount) > $balance) {
            $this->log("Cash Escrow Total is bigger than current balance", $transactionInfo);
            $key = $this->getEscrowRedisKey($userId);
            $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->set($key, $balance, BalanceEnum::EX, BalanceEnum::EXPIRATION);
            return $this->getEscrowBalance($userId);
        }
        $this->log("Add Cash Escrow", $transactionInfo);
        return $this->adjustEscrowBalance($transactionInfo);
    }

    /**
     * Deduct Balance to user on Redis
     *
     * @param $userId
     * @param $amount
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     * @throws InvalidBalanceException
     */
    public function deductBalance($userId, $amount, array $information = [])
    {
        $this->validateAmount($amount);
        $transactionInfo = new TransactionInfoVO($userId, -$amount, $information);
        return $this->adjustBalance($transactionInfo);
    }

    /**
     * Deduct EscrowBalance to user on Redis
     *
     * @param $userId
     * @param $amount
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     */
    public function deductEscrowBalance($userId, $amount, array $information = [])
    {
        $this->validateAmount($amount);
        $transactionInfo = new TransactionInfoVO($userId, -$amount, $information);
        $escrowBalance = $transactionInfo->setEscrowBalance($this->getEscrowBalance($userId));
        if ($escrowBalance >= $amount) {
            $this->log("Deduct Cash Escrow", $transactionInfo);
            return $this->adjustEscrowBalance($transactionInfo);
        }
        $this->log("Cash Escrow Balance is smaller than amount requested to deduct", $transactionInfo);
        $key = $this->getEscrowRedisKey($userId);
        $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->set($key, 0, BalanceEnum::EX, BalanceEnum::EXPIRATION);
        return $this->roundBalance(0);
    }

    /**
     * Either Deduct or Add Balance on Redis + add transaction for the balance change
     *
     * @param TransactionInfoVO $transactionInfo
     * @return float
     * @throws InvalidBalanceException
     */
    protected function adjustBalance(TransactionInfoVO $transactionInfo)
    {
        $userId = $transactionInfo->getUserId();
        $amount = $transactionInfo->getAmount();

        $transactionInfo->setBalance($this->getBalance($userId));
        $newBalance = $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->incrbyfloat($this->getBalanceRedisKey($userId), $amount);
        $newBalance = $this->roundBalance($newBalance);
        if ($newBalance < 0) {
            $this->log('Cash Transactions Discrepancies', $transactionInfo);
            $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->incrbyfloat($this->getBalanceRedisKey($userId), -$amount);
            throw new InvalidBalanceException();
        }
        $oldBalance = $transactionInfo->setBalance($this->roundBalance($newBalance - $amount));
        $now = date("Y-m-d H:i:s");
        $transactionArchived = [
            'user_id' => $userId,
            'amount' => $amount,
            'old_cash_balance' => $oldBalance,
            'cash_balance' => $newBalance,
            'reason' => $transactionInfo->getReason(),
            'remarks' => $transactionInfo->getRemarks(),
            'created_at' => $now,
            'updated_at' => $now
        ];
        $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->rpush($this->getTransactionsRedisKey($userId), json_encode($transactionArchived));
        $this->log('Cash Transactions', $transactionInfo);
        return $newBalance;
    }

    /**
     * Either Deduct or Add EscrowBalance on Redis + add transaction for the balance change
     *
     * @param TransactionInfoVO $transactionInfo
     * @return float
     */
    protected function adjustEscrowBalance(TransactionInfoVO $transactionInfo)
    {
        $key = $this->getEscrowRedisKey($transactionInfo->getUserId());
        $balance = $transactionInfo->setBalance($this->getEscrowBalanceFromCache($key));
        if ($balance == 0) {
            $this->log("Cash Escrow Balance not found", $transactionInfo);
        }
        $newEscrowBalance = $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->incrbyfloat($key, $transactionInfo->getAmount());
        $newEscrowBalance = $transactionInfo->setNewBalance($this->roundBalance($newEscrowBalance));
        $this->log("Cash Escrow Balance updated", $transactionInfo);
        return $newEscrowBalance;
    }

    /**
     * @param $userId
     * @param $amount
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     * @throws InvalidBalanceException
     */
    public function setBalance($userId, $amount, array $information = [])
    {
        return $this->processSetBalance($this->getBalance($userId), $amount, $userId, $information);
    }

    /**
     * @param $userId
     * @param $amount
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     * @throws InvalidBalanceException
     */
    public function setEscrowBalance($userId, $amount, array $information = [])
    {
        return $this->processSetBalance($this->getEscrowBalance($userId), $amount, $userId, $information);
    }

    /**
     * Set the balance. Instead of directly assigning the value to redis, the difference in amount is added/deducted and maintain the transactions
     *
     * @param $balance
     * @param $amount
     * @param $userId
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     * @throws InvalidBalanceException
     */
    private function processSetBalance($balance, $amount, $userId, array $information): float
    {
        $this->validateAmount($amount);
        $transactionInfo = new TransactionInfoVO($userId, $amount, $information);
        $amount = abs($amount);
        $transactionInfo->setAmount($amount - $balance);
        return $this->adjustBalance($transactionInfo);
    }

    /**
     * Get the balance from redis. If redis doesn't exist, adjust the balance from transactions and store to redis
     *
     * @return float
     */
    public function getBalance($userId)
    {
        $balance = $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->get($this->getBalanceRedisKey($userId));
        if (!$balance) {
            if ($userId) {
                $balance = $this->getBalanceFromDb($userId);
            } else {
                $balance = 0;
            }
        }
        return $this->roundBalance($balance);
    }

    /**
     * Get the escrow balance from redis.
     *
     * @return float
     */
    public function getEscrowBalance($userId)
    {
        $key = $this->getEscrowRedisKey($userId);
        $balance = $this->getEscrowBalanceFromCache($key);
        return $this->roundBalance($balance);
    }

    /**
     * Get the balance from DB
     *
     * @param $userId
     * @return float|int
     */
    protected function getBalanceFromDb($userId)
    {
        $transactionInfoVO = new TransactionInfoVO($userId, 0);
        /** !!! IMPORTANT Don't use Eloquent ---> loop when resolving getCashBalanceAttribute() */
        $user = $this->db::connection($this->getConnectionName())->table(WalletEnum::USERS_TABLE)
            ->select(WalletEnum::USER_ID, WalletEnum::CASH_BALANCE)
            ->where(WalletEnum::USER_ID, $userId)
            ->first();
        if (!$user) {
            $this->log( 'User not found for cash_balance from DB. Setting balance to 0', $transactionInfoVO);
            return 0;
        }
        $transactionInfoVO->setBalance($user->cash_balance);
        $this->log( 'Fetch user cash_balance from DB', $transactionInfoVO);

        $count = $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->llen($this->getTransactionsRedisKey($userId));
        $transactionInfoVO->setCountTransactions($count);
        $this->log('Count transactions of user from Redis with Key Transactions:' . $userId, $transactionInfoVO);

        $transactions = $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->lrange($this->getTransactionsRedisKey($userId), 0, $count);
        $this->log('Fetch now transactions of user from Redis with Key Transactions:' . $userId, $transactionInfoVO);
        $sum = 0;
        foreach ($transactions as $transaction) {
            $sum += json_decode($transaction)->amount;
        }
        $newBalance = $transactionInfoVO->setNewBalance($user->cash_balance + $sum);
        $this->log('Final balance of the user using Transactions from Redis', $transactionInfoVO);
        $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->set($this->getBalanceRedisKey($userId), $newBalance);
        $this->log('Saved final balance of the user on Redis', $transactionInfoVO);
        return $this->roundBalance($newBalance);
    }

    /**
     * @param string $key
     * @return int
     */
    protected function getEscrowBalanceFromCache(string $key)
    {
        $balance = $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->get($key);
        //set to is_null because balance can be 0 which is an actual value and get exist
        if (is_null($balance)) {
            $this->redis::connection(BalanceEnum::REDIS_CONNECTION)->set($key, 0, BalanceEnum::EX, BalanceEnum::EXPIRATION);
            $balance = 0;
        }
        return $balance;
    }

    /**
     * @return string
     */
    protected function getConnectionName()
    {
        return WalletEnum::AUTH_CONNECTION;
    }

    /**
     * @param $amount
     * @throws InvalidAmountException
     */
    private function validateAmount($amount)
    {
        if (is_null($amount) || is_bool($amount)) {
            throw new InvalidAmountException();
        }
    }

    /**
     * @param $balance
     * @return float
     */
    private function roundBalance($balance)
    {
        return round($balance, BalanceEnum::PRECISION);
    }

    /**
     * @param $userId
     * @return string
     */
    private function getBalanceRedisKey($userId)
    {
        return "User:$userId:balance";
    }

    /**
     * @param $userId
     * @return string
     */
    private function getEscrowRedisKey($userId)
    {
        return "User:$userId:escrow";
    }

    /**
     * @param $userId
     * @return string
     */
    private function getTransactionsRedisKey($userId)
    {
        return "Transactions:$userId";
    }
}
